﻿using System;
using System.Diagnostics;
using System.IO;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Threading;

namespace TVAppLauncher
{
    public partial class App
    {
        private Semaphore _ApplicationInstanceSemaphore;

        protected override void OnStartup(StartupEventArgs e)
        {
            base.OnStartup(e);

            bool createdNew;
            // Using the semaphore to prevent multiple instances from running.
            _ApplicationInstanceSemaphore = new Semaphore(1, 2, "TVAppLauncherApplicationSemaphore", out createdNew);

            var args = new StartupArguments(e.Args);
            var windowTitle = GetWindowTitle(args.Url, args.Exe);

            if (!createdNew)
            {
                var existingWindow = NativeMethods.FindWindowEx(IntPtr.Zero, IntPtr.Zero, null, windowTitle);
                if (existingWindow != IntPtr.Zero)
                {
                    SendMessageAndQuit(existingWindow, args.Quit);
                    return;
                }
            }

            if (args.Quit && !args.Start)
                Shutdown();
            else
                Start(args, windowTitle);
        }

        private async void SendMessageAndQuit(IntPtr existingWindow, bool quit)
        {
            var message = quit ? CustomWindowMessage.Quit : CustomWindowMessage.BringToTop;
            NativeMethods.PostMessage(existingWindow, (WindowMessage)message, IntPtr.Zero, IntPtr.Zero);
            await Task.Delay(3000);
            Shutdown();
        }

        private async void Start(StartupArguments arguments, string title)
        {
            var isBrowser = IsBrowser(arguments.ExeId);
            var showOpeningBackdrop = !arguments.NoOpeningOverlay && (arguments.Url != null || isBrowser);

            var mainWindow = new MainWindow
            {
                Title = title,
                ShowOpeningOverlay = showOpeningBackdrop,
                InputProcessor = arguments.InputProcessor,
                HideCursor = true
            };

            await EventAsync.FromRoutedEvent(
                h => mainWindow.Loaded += h,
                h => mainWindow.Loaded -= h,
                mainWindow.Show);

            NativeMethods.HideCursor(mainWindow.ThisWindowHandle);

            // Delay starting process to prevent possible white flicker when starting browser.
            if (isBrowser && showOpeningBackdrop)
                await Task.Delay(300);

            var process = await StartProcess(arguments, mainWindow);
            if (process == null)
                return;

            mainWindow.AttachedProcess = process;

            if (arguments.ForceFullScreen)
            {
                NativeMethods.MakeFullScreen(process.MainWindowHandle);
            }
            else if (arguments.ExeId == KnownExeIdentifier.Vivaldi)
            {
                // The start-fullscreen switch doesn't work for vivaldi.
                KeyboardHelper.SendKeyDownKeyUp(process.MainWindowHandle, VirtualKeyCode.F11, 0x57, false);
            }

            if (mainWindow.ShowOpeningOverlay)
                mainWindow.HideOpeningOverlay(1700);
        }

        private async Task<Process> StartProcess(StartupArguments arguments, MainWindow mainWindow)
        {
            var process = new Process
            {
                StartInfo = GetProcessStartInfo(arguments, arguments.ExeId),
                EnableRaisingEvents = true
            };

            process.Exited += async delegate
            {
                // If main window closed the process, give it time to complete any animations before shutting down.
                await Task.Delay(1000);
                // Need to dispatch this because the event is raised on a different thread.
                var z = Dispatcher.BeginInvoke(DispatcherPriority.Send, (Action)Shutdown);
            };
            
            try
            {
                if (!process.Start())
                {
                    ShowErrorMessageAndQuit(mainWindow, $"Could not start process.\n\n{arguments.Exe}");
                    return null;
                }
            }
            catch (Exception e)
            {
                var error = $"{arguments.Exe} {arguments.Args ?? arguments.Url}\n\nError: {e.Message}";
                ShowErrorMessageAndQuit(mainWindow, error);
                return null;
            }

            var count = 0;
            while (process.MainWindowHandle == IntPtr.Zero && count++ < 300)
                await Task.Delay(5);

            if (count >= 300)
            {
                ShowErrorMessageAndQuit(mainWindow, "Could not obtain a handle to the process window;");
                return null;
            }

            return process;
        }

        private void ShowErrorMessageAndQuit(MainWindow mainWindow, string message)
        {
            // TODO: It would be better to pass this string to main window and let it display the error.
            // As it is now the user will have to use the mouse to close the message box.
            MessageBox.Show(mainWindow, message, "TVAppLauncher", MessageBoxButton.OK, MessageBoxImage.Error);
            Shutdown();
        }

        private static string GetWindowTitle(string url, string exe)
        {
            if (url != null)
            {
                if (url.StartsWith("https://"))
                    return url.Remove(0, 8);
                if (url.StartsWith("http://"))
                    return url.Remove(0, 7);
                return url;
            }

            if (exe != null)
            {
                if (File.Exists(exe))
                    return Path.GetFileNameWithoutExtension(exe);
                return exe;
            }

            return "JRTV App Launcher";
        }

        private static bool IsBrowser(KnownExeIdentifier knownExe)
        {
            return knownExe == KnownExeIdentifier.Chrome || knownExe == KnownExeIdentifier.Firefox || knownExe == KnownExeIdentifier.Vivaldi;
        }

        private static ProcessStartInfo GetProcessStartInfo(StartupArguments args, KnownExeIdentifier knownExe)
        {
            var startInfo = new ProcessStartInfo
            {
                FileName = args.Exe,
                Arguments = args.Args,
                WindowStyle = ProcessWindowStyle.Maximized,
            };

            if (args.Args != null || args.Url == null)
                return startInfo;

            switch (knownExe)
            {
                case KnownExeIdentifier.Chrome:
                    startInfo.Arguments = GetChromeStartupArgs(args.Url);
                    break;
                case KnownExeIdentifier.Firefox:
                    startInfo.Arguments = GetFirefoxStartupArgs(args.Url);
                    break;
                case KnownExeIdentifier.Vivaldi:
                    startInfo.Arguments = GetVivalidStartupArgs(args.Url);
                    break;
                default:
                    startInfo.Arguments = args.Url;
                    break;
            }

            return startInfo;
        }

        private static string GetChromeStartupArgs(string url)
        {
            // Including user data dir in the command allows the new window to go into fullscreen mode when 
            // there is an existing instance of Chrome already running.
            var userDataDir = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData), "Google", "Chrome", "YouTubeLauncherData");
            return $"--new-window --no-first-run --no-default-browser-check --start-maximized --start-fullscreen --user-data-dir={userDataDir} {url}"; // 
        }

        private static string GetVivalidStartupArgs(string url)
        {
            var userDataDir = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData), "Vivaldi", "YouTubeLauncherData");
            return $"--new-window --no-first-run --no-default-browser-check --start-maximized --user-data-dir={userDataDir} {url}";
        }

        private static string GetFirefoxStartupArgs(string url)
        {
            return $"-new-window {url}";
        }
    }
}
