﻿using System;
using System.Threading.Tasks;
using System.Windows;

namespace TVAppLauncher
{
    // Based on: http://social.msdn.microsoft.com/Forums/sk/async/thread/30f3339c-5e04-4aa8-9a09-9be72d9d9a1b
    public static class EventAsync
    {
        /// <summary>
        /// Creates a Task that waits for an event to occur, where the event's specified handler type is EventHandler.
        /// </summary>
        /// <example>
        /// <![CDATA[
        /// await EventAsync.FromEvent(
        ///     eh => instance.SomeEvent += eh,
        ///     eh => instance.SomeEvent -= eh,
        ///     storyboard.Begin);
        /// ]]>
        /// </example>
        public static Task FromEvent(Action<EventHandler> addEventHandler, Action<EventHandler> removeEventHandler, Action beginAction = null)
        {
            var taskCompletionSource = new TaskCompletionSource<object>();

            void handler(object sender, EventArgs e)
            {
                removeEventHandler(handler);
                taskCompletionSource.SetResult(null);
            }

            addEventHandler(handler);
            beginAction?.Invoke();

            return (Task)taskCompletionSource.Task;
        }

        /// <summary>
        /// Creates a Task that waits for an event to occur, where the even't specified handler type is EventHandler&lt;T>
        /// </summary>
        /// <example>
        /// <![CDATA[
        /// await EventAsync.FromEvent<object>(
        ///     eh => storyboard.Completed += eh,
        ///     eh => storyboard.Completed -= eh,
        ///     storyboard.Begin);
        /// ]]>
        /// </example>
        public static Task FromEvent<T>(Action<EventHandler<T>> addEventHandler, Action<EventHandler<T>> removeEventHandler, Action beginAction = null)
        {
            var taskCompletionSource = new TaskCompletionSource<object>();

            void handler(object sender, T e)
            {
                removeEventHandler(handler);
                taskCompletionSource.SetResult(null);
            }

            addEventHandler(handler);
            beginAction?.Invoke();

            return (Task)taskCompletionSource.Task;
        }

        /// <summary>
        /// Creates a Task
        /// that waits for an event to occur.
        /// </summary>
        /// <example>
        /// <![CDATA[
        /// await EventAsync.FromEvent(
        ///     eh => button.Click += eh,
        ///     eh => button.Click -= eh);
        /// ]]>
        /// </example>
        public static Task<RoutedEventArgs> FromRoutedEvent(
            Action<RoutedEventHandler> addEventHandler,
            Action<RoutedEventHandler> removeEventHandler,
            Action beginAction = null)
        {
            var taskCompletionSource = new TaskCompletionSource<RoutedEventArgs>();

            void handler(object s, RoutedEventArgs e)
            {
                removeEventHandler(handler);
                taskCompletionSource.SetResult(e);
            }

            addEventHandler(handler);
            beginAction?.Invoke();

            return taskCompletionSource.Task;
        }
    }
}
