﻿using System;
using System.Runtime.InteropServices;

namespace TVAppLauncher
{
    internal enum MCERemoteButtonCode
    {
        Invalid = 0,
        GreenButton = 0x0D,
        DVDMenu = 0x24,
        LiveTV = 0x25,
        RecordedTV = 0x48,
        Guide = 0x8D,
        ChannelUp = 0x9C,
        ChannelDown = 0x9D,
        Play = 0xB0,
        Pause = 0xB1,
        Record = 0xB2,
        FastForward = 0xB3,
        Rewind = 0xB4,
        Skip = 0xB5,
        Previous = 0xB6,
        Stop = 0xB7,
        Mute = 0xE2,
        VolumeUp = 0xE9,
        VolumeDown = 0xEA,
        MoreDetails = 0x209,
        Back = 0x224,
    }

    internal static class MCERemoteHelper
    {
        /// <summary>
        /// Its necessary to register to recieve WM_INPUT events from HID devices.
        /// Certain buttons on the MCE remote control only send WM_INPUT message.
        /// </summary>
        public static void RegisterMCERemoteControl(IntPtr windowHandle)
        {
            var rid = new RawInputDevice[4];

            rid[0].UsagePage = HidUsagePage.Consumer;
            rid[0].Usage = HidUsage.Pointer;
            rid[0].Flags = RawInputDeviceFlags.InputSink;
            rid[0].WindowHandle = windowHandle;

            rid[1].UsagePage = HidUsagePage.VendorDefined;
            rid[1].Usage = HidUsage.VendorDefined;
            rid[1].Flags = RawInputDeviceFlags.InputSink;
            rid[1].WindowHandle = windowHandle;

            RegisterRawInputDevices(rid, 2, Marshal.SizeOf(rid[0]));
        }

        /// <summary>
        /// Retrieves the data from a WM_INPUT message used to identify which button was pressed.
        /// </summary>
        public static MCERemoteButtonCode GetMCEButtonCode(IntPtr lParam)
        {
            RawInput input;
            var size = Marshal.SizeOf(typeof(RawInput));
            var outSize = GetRawInputData(lParam, RawInputCommand.Input, out input, ref size, Marshal.SizeOf(typeof(RawInputHeader)));

            if (outSize == -1)
                return MCERemoteButtonCode.Invalid;

            if (input.Header.Type == RawInputType.Hid)
            {
                var rawData = (int)input.RawData.Hid.Data;
                // I don't understand why, but this shift is necessary to obtain the correct value from the raw data.
                // The field Hid.Data is a byte array. Or a pointer to an array I'd guess.
                // But I don't know how obtain an actual byte array from the raw data value.
                return (MCERemoteButtonCode)(rawData >> 8);
            }

            return MCERemoteButtonCode.Invalid;
        }

        [DllImport("user32.dll")]
        private static extern bool RegisterRawInputDevices([MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 0)] RawInputDevice[] pRawInputDevices, int uiNumDevices, int cbSize);

        [DllImport("user32.dll")]
        private static extern int GetRawInputData(IntPtr hRawInput, RawInputCommand uiCommand, out RawInput pData, ref int pcbSize, int cbSizeHeader);

        [StructLayout(LayoutKind.Sequential)]
        private struct RawInputDevice
        {
            /// <summary>Top level collection Usage page for the raw input device.</summary>
            public HidUsagePage UsagePage;
            /// <summary>Top level collection Usage for the raw input device. </summary>
            public HidUsage Usage;
            /// <summary>Mode flag that specifies how to interpret the information provided by UsagePage and Usage.</summary>
            public RawInputDeviceFlags Flags;
            /// <summary>Handle to the target device. If NULL, it follows the keyboard focus.</summary>
            public IntPtr WindowHandle;
        }

        [Flags]
        private enum RawInputDeviceFlags
        {
            /// <summary>No flags.</summary>
            None = 0,
            /// <summary>If set, this removes the top level collection from the inclusion list. 
            /// This tells the operating system to stop reading from a device which matches the top level collection.</summary>
            Remove = 0x00000001,
            /// <summary>If set, this specifies the top level collections to exclude when reading a complete usage page. 
            /// This flag only affects a TLC whose usage page is already specified with PageOnly.</summary>
            Exclude = 0x00000010,
            /// <summary>If set, this specifies all devices whose top level collection is from the specified usUsagePage. 
            /// Note that Usage must be zero. To exclude a particular top level collection, use Exclude.</summary>
            PageOnly = 0x00000020,
            /// <summary>If set, this prevents any devices specified by UsagePage or Usage from generating legacy messages. 
            /// This is only for the mouse and keyboard.</summary>
            NoLegacy = 0x00000030,
            /// <summary>If set, this enables the caller to receive the input even when the caller is not in the foreground. 
            /// Note that WindowHandle must be specified.</summary>
            InputSink = 0x00000100,
            /// <summary>If set, the mouse button click does not activate the other window.</summary>
            CaptureMouse = 0x00000200,
            /// <summary>If set, the application-defined keyboard device hotkeys are not handled. 
            /// However, the system hotkeys; for example, ALT+TAB and CTRL+ALT+DEL, are still handled. 
            /// By default, all keyboard hotkeys are handled. NoHotKeys can be specified even if NoLegacy is not specified and WindowHandle is NULL.</summary>
            NoHotKeys = 0x00000200,
            /// <summary>If set, application keys are handled.  NoLegacy must be specified.  Keyboard only.</summary>
            AppKeys = 0x00000400
        }


        private enum HidUsagePage : ushort
        {
            Generic = 0x01,
            Consumer = 0x0C,
            Keyboard = 0x07,
            VendorDefined = 0xFFBC,
        }

        private enum HidUsage : ushort
        {
            Pointer = 0x01,
            Mouse = 0x02,
            Keyboard = 0x06,
            VendorDefined = 0x88,
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct RawInput
        {
            public RawInputHeader Header;
            public RawData RawData;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct RawInputHeader
        {
            /// <summary>Type of device the input is coming from.</summary>
            public RawInputType Type;
            /// <summary>Size of the packet of data.</summary>
            public int Size;
            /// <summary>Handle to the device sending the data.</summary>
            public IntPtr Device;
            /// <summary>wParam from the window message.</summary>
            public IntPtr wParam;
        }

        private enum RawInputType
        {
            Mouse = 0,
            Keyboard = 1,
            Hid = 2,
            Other = 3
        }

        private enum RawInputCommand
        {
            Input = 0x10000003,
            Header = 0x10000005
        }

        [StructLayout(LayoutKind.Explicit)]
        private struct RawData
        {
            //[FieldOffset(0)]
            //internal Rawmouse mouse;
            [FieldOffset(0)]
            public RawInputKeyboard Keyboard;
            [FieldOffset(0)]
            public RawInputHid Hid;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct RawInputHid
        {
            /// <summary>Size of the Hid data in bytes.</summary>
            public int Size;
            /// <summary>Number of Hid in Data.</summary>
            public int Count;
            /// <summary>Data for the Hid.</summary>
            public IntPtr Data;
        }

        private struct RawInputKeyboard
        {
            public ushort Makecode;                 // Scan code from the key depression
            public ushort Flags;                    // One or more of RI_KEY_MAKE, RI_KEY_BREAK, RI_KEY_E0, RI_KEY_E1
            public ushort Reserved;                 // Always 0    
            public ushort VKey;                     // Virtual Key Code
            public uint Message;                    // Corresponding Windows message for exmaple (WM_KEYDOWN, WM_SYASKEYDOWN etc)
            public uint ExtraInformation;           // The device-specific addition information for the event (seems to always be zero for keyboards)            
        }
    }
}
