﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Interop;
using System.Windows.Media.Animation;
using System.Windows.Threading;

namespace TVAppLauncher
{
    public partial class MainWindow
    {
        #region Fields, Construction

        private GlobalKeyboardHook _GlobalKeyboardHook;

        internal MainWindow()
        {
            InitializeComponent();
            Topmost = true;
        }

        internal IntPtr ThisWindowHandle { get; private set; }
        internal IntPtr AttachedWindowHandle => AttachedProcess?.MainWindowHandle ?? IntPtr.Zero;

        internal bool HideCursor { get; set; }
        internal IInputProcessor InputProcessor { get; set; }
        internal bool ShowOpeningOverlay { get; set; }
        internal Process AttachedProcess { get; set; }


        #endregion // Fields, Construction

        #region Overrides

        protected override void OnSourceInitialized(EventArgs e)
        {
            base.OnSourceInitialized(e);

            OpeningOverlay.Visibility = ShowOpeningOverlay ? Visibility.Visible : Visibility.Collapsed;

            ThisWindowHandle = new WindowInteropHelper(this).Handle;
            // Makes it so the window doesn't show up when user cycles
            // through desktop windows with Alt-Tab
            NativeMethods.MakeToolWindowStyle(ThisWindowHandle, true);

            // Create a global keyboard hook and register for its KeyEvent event.
            _GlobalKeyboardHook = new GlobalKeyboardHook();
            _GlobalKeyboardHook.KeyEvent += GlobalKeyboardHook_KeyEvent;

            // Register to recieve WM_INPUT messages from the MCE remote control. 
            // These messages are recieved in the windows message loop hook.
            MCERemoteHelper.RegisterMCERemoteControl(ThisWindowHandle);

            // Add a hook to the windows message loop.
            HwndSource.FromHwnd(ThisWindowHandle)?.AddHook(WinProc);
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            base.OnClosing(e);

            CloseAttachedProcess();

            _GlobalKeyboardHook?.Dispose();
            var hwndSource = HwndSource.FromHwnd(ThisWindowHandle);
            hwndSource?.RemoveHook(WinProc);
        }

        protected override void OnDeactivated(EventArgs e)
        {
            base.OnDeactivated(e);
            if (IsMenuVisible)
                HideMenu();
        }

        #endregion // Overrides

        #region Window Management

        private void CloseAttachedProcess()
        {
            if (AttachedProcess == null)
                return;
            try
            {
                AttachedProcess.CloseMainWindow();
                AttachedProcess.Close();
                AttachedProcess = null;
            }
            catch { }
        }

        internal async void DelayClose()
        {
            await Task.Delay(30);
            CloseAttachedProcess();
            Close();
        }

        internal async void HideOpeningOverlay(int delay)
        {
            if (delay > 0)
                await Task.Delay(delay);

            var storyboard = new Storyboard();
            var opacityAnimation = new DoubleAnimation
            {
                From = 1.0,
                To = 0.0,
                Duration = TimeSpan.FromSeconds(1),
                EasingFunction = new CircleEase { EasingMode = EasingMode.EaseIn }
            };

            storyboard.Children.Add(opacityAnimation);
            Storyboard.SetTarget(opacityAnimation, OpeningOverlay);
            Storyboard.SetTargetProperty(opacityAnimation, new PropertyPath(OpacityProperty));

            storyboard.Completed += delegate
            {
                OpeningOverlay.Visibility = Visibility.Collapsed;
                ShowWindow();
            };

            storyboard.Begin(this);
        }

        internal bool IsMenuVisible => ExitMenu.Visibility == Visibility.Visible;

        internal void ShowMenu()
        {
            if (ExitMenu.Visibility == Visibility.Visible)
                return;
            WindowState = WindowState.Maximized;
            ExitMenu.Visibility = Visibility.Visible;
            ExitButton.Focus();
            _IsVisible = true;
        }

        private void HideMenu()
        {
            ExitMenu.Visibility = Visibility.Collapsed;
            WindowState = WindowState.Minimized;
        }

        private async void ShowAttachedProcessWindow()
        {
            if (AttachedWindowHandle == IntPtr.Zero)
                return;

            HideMenu();

            NativeMethods.SetWindowPos(AttachedWindowHandle, SetWindowPosType.Top, SetWindowPosFlag.ShowWindow);
            NativeMethods.ShowWindow(AttachedWindowHandle, ShowWindowFlag.Maximize);

            var isForeground = NativeMethods.SetForegroundWindow(AttachedWindowHandle);
            if (isForeground)
                return;

            if (GetActiveWindow(out isForeground) == AttachedWindowHandle)
                return;

            NativeMethods.SetWindowPlacement(AttachedWindowHandle, SetWindowPosType.TopMost);
            await Task.Delay(500);
            NativeMethods.SetWindowPlacement(AttachedWindowHandle, SetWindowPosType.NoTopMost);
            NativeMethods.SetWindowPlacement(AttachedWindowHandle, SetWindowPosType.Top);

            _IsVisible = true;
        }

        private async void ShowWindow(int delay = 0)
        {
            if (delay > 0)
                await Task.Delay(delay);
            if (AttachedWindowHandle != IntPtr.Zero)
                ShowAttachedProcessWindow();
            else
                Activate();
            _IsVisible = true;
        }

        private void HideWindow()
        {
            if (AttachedWindowHandle != IntPtr.Zero)
            {
                NativeMethods.SetWindowPos(AttachedWindowHandle, SetWindowPosType.Bottom, SetWindowPosFlag.NoActivate);
                if (NativeMethods.GetForegroundWindow() == AttachedWindowHandle)
                    NativeMethods.ShowWindow(AttachedWindowHandle, ShowWindowFlag.Minimize);
            }

            HideMenu();
            _IsVisible = false;
        }

        private bool _HasHadForeground;
        private bool _IsVisible;

        private IntPtr GetActiveWindow(out bool hasForeground)
        {
            //Debug.WriteLine($"Foreground Window: {NativeMethods.GetWindowText(foregroundWindow)} - Current Window: {NativeMethods.GetWindowText(activeWindow)}");
            IntPtr activeWindow;
            if (IsMenuVisible || AttachedWindowHandle == IntPtr.Zero)
                activeWindow = ThisWindowHandle;
            else
                activeWindow = AttachedWindowHandle;

            var foregroundWindow = NativeMethods.GetForegroundWindow();
            if (activeWindow == foregroundWindow)
                hasForeground = _HasHadForeground = true;
            else
                hasForeground = false;

            _HasHadForeground |= hasForeground;
            _IsVisible = hasForeground || NativeMethods.IsWindowOnTop(activeWindow);

            return _IsVisible ? activeWindow : IntPtr.Zero;
        }

        #endregion Window Management

        #region UIElement Event Handlers

        private void CancelButton_Click(object sender, RoutedEventArgs e)
        {
            ShowWindow();
        }

        private void ExitButton_Click(object sender, RoutedEventArgs e)
        {
            DelayClose();
        }

        private void HideButton_Click(object sender, RoutedEventArgs e)
        {
            HideWindow();
        }

        #endregion

        #region Keyboard Event Processing

        /// <summary>
        /// This is the event handler for the KeyEvent event of the GlobalKeyboardHook class instance.        
        /// </summary>
        private void GlobalKeyboardHook_KeyEvent(object sender, KeyEventArgs e)
        {
            bool hasForeground;

            var activeWindow = GetActiveWindow(out hasForeground);
            if (activeWindow == IntPtr.Zero)
                return;

            if (InputProcessor?.ProcessGlobalKeyboardEvent(this, e, activeWindow, hasForeground) == true)
                return;

            // Don't need to process system keys. (Alt key is sent with messages SysKeyDown and KeyUp...)
            if (e.VirtualKeyCode == VirtualKeyCode.Alt || e.Message == WindowMessage.SysKeyDown || e.Message == WindowMessage.SysKeyUp)
                return;

            // Nullable for three-state. The null value means send both key up and key down.
            bool? sendKey = false;

            switch (e.VirtualKeyCode)
            {
                case VirtualKeyCode.F4:
                    if (e.AltDown)
                    {
                        e.Handled = true;
                        DelayClose();
                    }
                    break;
                case VirtualKeyCode.BrowserBack:
                    e.Handled = true;
                    if (e.Message == WindowMessage.KeyDown)
                        LongPressHelper.RegisterPressed(ShowMenu);
                    else if (LongPressHelper.RegisterReleased())
                        sendKey = null;
                    break;
                case VirtualKeyCode.Stop:
                    e.Handled = true;
                    if (e.Message == WindowMessage.KeyDown)
                        LongPressHelper.RegisterPressed(ShowMenu);
                    else if (LongPressHelper.RegisterReleased())
                        sendKey = null;
                    break;
                case VirtualKeyCode.Escape:
                    e.Handled = true;
                    if (e.Message == WindowMessage.KeyDown)
                        LongPressHelper.RegisterPressed(DelayClose);
                    else if (LongPressHelper.RegisterReleased())
                        sendKey = null;
                    break;
                default:
                    // Prevent JRMC from processing keys if it has focus.
                    sendKey = e.Handled = !hasForeground;
                    break;
            }

            if (sendKey == false)
                return;

            var keyCode = e.VirtualKeyCode;
            var scanCode = e.ScanCode;
            var extended = e.Extended;

            InputProcessor?.ReplaceKeyIfNeeded(ref keyCode, ref scanCode, ref extended);

            if (!sendKey.HasValue)
                KeyboardHelper.SendKeyDownKeyUp(activeWindow, keyCode, scanCode, extended);
            else if (e.Message == WindowMessage.KeyUp)
                KeyboardHelper.SendKeyUp(activeWindow, keyCode, scanCode, extended);
            else
                KeyboardHelper.SendKeyDown(activeWindow, keyCode, scanCode, extended);
        }

        #endregion // Keyboard

        #region Windows Message Loop

        private IntPtr WinProc(IntPtr hwnd, int msg, IntPtr wParam, IntPtr lParam, ref bool handled)
        {
            switch (msg)
            {
                case (int)CustomWindowMessage.BringToTop:
                    ShowWindow();
                    handled = true;
                    break;
                case (int)CustomWindowMessage.Quit:
                    Close();
                    handled = true;
                    break;
                case (int)WindowMessage.Input:
                    ProcessWMInputMessage(lParam, out handled);
                    break;
            }

            return IntPtr.Zero;
        }

        // The remote control sends WM_INPUT message for certain buttons.
        // It is necessary to register to recieve these messages using the windows
        // API method RegisterRawInputDevices. See RawInputDeviceHelper for implementation.
        // See https://msdn.microsoft.com/en-us/library/windows/desktop/bb417079.aspx for details.
        private void ProcessWMInputMessage(IntPtr lParam, out bool handled)
        {
            var wasVisible = _IsVisible;

            bool hasForeground;
            var activeWindow = GetActiveWindow(out hasForeground);

            var code = MCERemoteHelper.GetMCEButtonCode(lParam);
            handled = InputProcessor?.ProcessMCEButtonPress(this, code, activeWindow, hasForeground) == true;

            if (!_IsVisible && (_HasHadForeground || !wasVisible))
                return;

            switch (code)
            {
                case MCERemoteButtonCode.GreenButton:
                    if (!handled)
                    {
                        handled = true;
                        DelayClose();
                    }
                    break;
                case MCERemoteButtonCode.Play:
                case MCERemoteButtonCode.FastForward:
                case MCERemoteButtonCode.Rewind:
                case MCERemoteButtonCode.Skip:
                case MCERemoteButtonCode.Previous:
                case MCERemoteButtonCode.Pause:
                    if (!hasForeground && _IsVisible)
                    {
                        handled = true;
                        WindowPlacementHelper.RegisterProblematicInput(activeWindow);
                    }
                    break;
            }
        }

        #endregion

        #region Window Placement Helper

        private static class WindowPlacementHelper
        {
            private static IntPtr _HWnd = IntPtr.Zero;
            private static DispatcherTimer _Timer;

            // Its not possible to prevent JRiver from recieving the WM_INPUT messages sent by the remote control.
            // Sometimes JR will bring itself to the front in response, when it shouldn't.
            // The only way to prevent it is to make the window Topmost temporarily.
            public static void RegisterProblematicInput(IntPtr hWnd)
            {
                if (_HWnd == IntPtr.Zero)
                {
                    _HWnd = hWnd;
                    NativeMethods.SetWindowPlacement(_HWnd, SetWindowPosType.TopMost);
                }

                if (_Timer == null)
                {
                    _Timer = new DispatcherTimer(TimeSpan.FromSeconds(2),
                        DispatcherPriority.ApplicationIdle, Timer_Tick,
                        Dispatcher.CurrentDispatcher);
                }
                else
                {
                    // Dispite what the documentation says, 
                    // calling Start alone doesn't reset the count down.
                    _Timer.Stop();
                }

                // Restart the timer everytime the button is pressed.
                _Timer.Start();
            }

            private static void Timer_Tick(object sender, EventArgs e)
            {
                _Timer.Stop();
                NativeMethods.SetWindowPlacement(_HWnd, SetWindowPosType.NoTopMost);
                _HWnd = IntPtr.Zero;
            }
        }

        #endregion
    }
}
