﻿using System;
using System.Runtime.InteropServices;
using System.Text;

namespace TVAppLauncher
{
    #region Enums

    internal enum CustomWindowMessage : uint
    {
        BringToTop = WindowMessage.User + 1,
        Quit = WindowMessage.User + 2,
    }

    internal enum WindowMessage : uint
    {
        Close = 0x0010,
        WindowPositionChanging = 0x0046,
        Input = 0x00FF,
        KeyDown = 0x0100,
        KeyUp = 0x0101,
        SysKeyDown = 0x0104,
        SysKeyUp = 0x0105,
        AppCommand = 0x0319,
        User = 0x0400,
    }

    internal enum SetWindowPosType
    {
        Bottom = 1,
        Top = 0,
        TopMost = -1,
        NoTopMost = -2,
    }

    [Flags]
    internal enum SetWindowPosFlag : uint
    {
        NoSize = 0x0001,
        NoMove = 0x0002,
        NoActivate = 0x0010,
        FrameChanged = 0x0020,
        ShowWindow = 0x0040,
        HideWindow = 0x0080,
    }

    [Flags]
    internal enum ShowWindowFlag
    {
        Hide = 0,
        Maximize = 3,
        Show = 5,
        Minimize = 6,
        Restore = 9,
    }

    internal enum AppCommand
    {
        BrowserBack = 1,
        BrowserForward = 2,
        VolumeMute = 8,
        VolumeDown = 9,
        VolumeUp = 10,
        MediaNextTrack = 11,
        MediaPreviousTrack = 12,
        MediaPause = 47,
        MediaPlay = 46,
    }

    #endregion // Enumerations

    internal class NativeMethods
    {
        #region Public Methods

        public static void SendAppCommand(IntPtr hWnd, AppCommand command)
        {
            SendMessage(hWnd, WindowMessage.AppCommand, hWnd, (IntPtr)((int)command * 0x10000));
        }

        public static bool SetWindowPos(IntPtr hWnd, SetWindowPosType type, SetWindowPosFlag flags)
        {
            return SetWindowPos(hWnd, (IntPtr)type, 0, 0, 0, 0, flags | SetWindowPosFlag.NoMove | SetWindowPosFlag.NoSize);
        }

        /// <summary>
        /// Makes it so the window isn't included when cycling through windows with Alt-Tab
        /// </summary>
        public static void MakeToolWindowStyle(IntPtr hWnd, bool makeToolWindow)
        {
            const int gwlExstyle = -20;
            const int wsExToolwindow = 0x00000080;

            var existingStyle = GetWindowLongPtr(hWnd, gwlExstyle).ToInt32();

            if (makeToolWindow)
                existingStyle |= wsExToolwindow;
            else
                existingStyle &= ~wsExToolwindow;

            SetWindowLongPtr(hWnd, gwlExstyle, (IntPtr)existingStyle);
        }

        public static void MakeFullScreen(IntPtr hWnd)
        {
            const int gwlStyle = -16;
            const int gwlExstyle = -20;
            const int wsCaption = 0x00C00000;
            const int wsThickFrame = 0x00040000;
            const int wsExWindowEdge = 0x00000100;
            const int wsExClientEdge = 0x00000200;
            const int wsExStaticEdge = 0x00020000;
            const int wsExDlgmodalframe = 0x00000001;
            const int monitorDefaultToNearest = 0x00000002;

            var style = GetWindowLongPtr(hWnd, gwlStyle).ToInt32();
            style &= ~(wsCaption | wsThickFrame);
            SetWindowLongPtr(hWnd, gwlStyle, (IntPtr)style);

            var exStyle = GetWindowLongPtr(hWnd, gwlExstyle).ToInt32();
            exStyle &= ~(wsExDlgmodalframe | wsExWindowEdge | wsExClientEdge | wsExStaticEdge);
            SetWindowLongPtr(hWnd, gwlExstyle, (IntPtr)exStyle);

            var hMon = MonitorFromWindow(hWnd, monitorDefaultToNearest);
            var monitorInfo = GetMonitorInfo(hMon);
            var rect = monitorInfo.rcMonitor;
            SetWindowPos(hWnd, IntPtr.Zero, rect.Left, rect.Top, rect.Right - rect.Left, rect.Bottom - rect.Top, SetWindowPosFlag.FrameChanged);
        }

        public static string GetWindowText(IntPtr hWnd)
        {
            int textLength = GetWindowTextLength(hWnd);
            var outText = new StringBuilder(textLength + 1);
            GetWindowText(hWnd, outText, outText.Capacity);
            return outText.ToString();
        }

        public static void SetWindowPlacement(IntPtr hWnd, SetWindowPosType type)
        {
            SetWindowPos(hWnd, (IntPtr)type, 0, 0, 0, 0, SetWindowPosFlag.NoMove | SetWindowPosFlag.NoSize);
        }

        /// <summary>
        /// Determines whether the window is visible, whether or not it has keyboard focus.
        /// </summary>
        public static bool IsWindowOnTop(IntPtr hWnd)
        {
            Rect rect;
            if (!GetWindowRect(hWnd, out rect))
                return false;
            Point center;
            center.X = (rect.Right - rect.Left) / 2;
            center.Y = (rect.Bottom - rect.Top) / 2;

            var testHwnd = WindowFromPoint(center);
            if (testHwnd == hWnd)
                return true;

            // GetRoot = 2
            var rootHwnd = GetAncestor(testHwnd, 2);
            return rootHwnd == hWnd;
        }

        public static void HideCursor(IntPtr hWnd)
        {
            Rect rect;
            if (GetWindowRect(hWnd, out rect))
                SetCursorPos(rect.Right, rect.Top + 200);
        }

        [DllImport("user32.dll")]
        [return: MarshalAs(UnmanagedType.Bool)]
        public static extern bool SetForegroundWindow(IntPtr hWnd);

        [DllImport("user32.dll")]
        public static extern IntPtr GetForegroundWindow();

        [DllImport("user32.dll", SetLastError = true, CharSet = CharSet.Auto)]
        public static extern IntPtr SendMessage(IntPtr windowHandle, WindowMessage msg, IntPtr wParam, IntPtr lParam);

        [DllImport("user32.dll", SetLastError = true, CharSet = CharSet.Auto)]
        public static extern IntPtr PostMessage(IntPtr windowHandle, WindowMessage msg, IntPtr wParam, IntPtr lParam);

        [DllImport("user32.dll", SetLastError = true)]
        public static extern IntPtr FindWindowEx(IntPtr parentHandle, IntPtr childAfter, string className, string windowTitle);

        [DllImport("user32.dll")]
        public static extern bool ShowWindow(IntPtr hWnd, ShowWindowFlag nCmdShow);

        [DllImport("user32.dll", SetLastError = true)]
        public static extern IntPtr SetActiveWindow(IntPtr hWnd);

        public delegate void WinEventDelegate(IntPtr hWinEventHook, uint eventType,
            IntPtr hwnd, int idObject, int idChild, uint dwEventThread, uint dwmsEventTime);

        [DllImport("user32.dll")]
        public static extern IntPtr SetWinEventHook(uint eventMin, uint eventMax, IntPtr
                hmodWinEventProc, WinEventDelegate lpfnWinEventProc, uint idProcess,
            uint idThread, uint dwFlags);


        #endregion // Public

        #region Private Methods

        [DllImport("user32.dll", EntryPoint = "GetMonitorInfo", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool _GetMonitorInfo(IntPtr hMonitor, [In, Out] MinitorInfo lpmi);

        [DllImport("user32.dll")]
        public static extern IntPtr MonitorFromWindow(IntPtr hwnd, uint dwFlags);

        [DllImport("user32.dll")]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool SetCursorPos(int x, int y);

        [DllImport("user32.dll", SetLastError = true)]
        private static extern bool SetWindowPos(IntPtr hWnd, IntPtr hWndInsertAfter, int x, int y, int cx, int cy, SetWindowPosFlag flags);

        [DllImport("user32.dll", SetLastError = true, CharSet = CharSet.Auto)]
        private static extern int GetWindowTextLength(IntPtr hWnd);

        [DllImport("user32.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        private static extern int GetWindowText(IntPtr hWnd, StringBuilder lpString, int nMaxCount);

        [DllImport("user32.dll", ExactSpelling = true)]
        private static extern IntPtr GetAncestor(IntPtr hwnd, int flags);

        [DllImport("user32.dll")]
        private static extern IntPtr WindowFromPoint(Point p);

        [DllImport("user32.dll", EntryPoint = "GetWindowLong", SetLastError = true)]
        private static extern int GetWindowLongPtr32(IntPtr hWnd, int nIndex);

        [DllImport("user32.dll", EntryPoint = "GetWindowLongPtr", SetLastError = true)]
        private static extern IntPtr GetWindowLongPtr64(IntPtr hWnd, int nIndex);

        // This is aliased as a macro in 32bit Windows.
        private static IntPtr GetWindowLongPtr(IntPtr hwnd, int nIndex)
        {
            var ret = 8 == IntPtr.Size ? GetWindowLongPtr64(hwnd, nIndex) : new IntPtr(GetWindowLongPtr32(hwnd, nIndex));
            return ret;
        }

        [DllImport("user32.dll", EntryPoint = "SetWindowLong", SetLastError = true)]
        private static extern int SetWindowLong32(IntPtr hWnd, int nIndex, int dwNewLong);

        [DllImport("user32.dll", EntryPoint = "SetWindowLongPtr", SetLastError = true)]
        private static extern IntPtr SetWindowLongPtr64(IntPtr hWnd, int nIndex, IntPtr dwNewLong);

        // Use instead of SetWindowLong to assure 64 bit compatibility.
        private static IntPtr SetWindowLongPtr(IntPtr hwnd, int nIndex, IntPtr dwNewLong)
        {
            if (8 == IntPtr.Size)
                return SetWindowLongPtr64(hwnd, nIndex, dwNewLong);
            return new IntPtr(SetWindowLong32(hwnd, nIndex, dwNewLong.ToInt32()));
        }

        [DllImport("user32.dll", SetLastError = true)]
        private static extern bool GetWindowRect(IntPtr hwnd, out Rect lpRect);

        #endregion // Private

        #region Types

        [StructLayout(LayoutKind.Sequential)]
        private struct Point
        {
            public int X;
            public int Y;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct Rect
        {
            public int Left;
            public int Top;
            public int Right;
            public int Bottom;
        }

        [StructLayout(LayoutKind.Sequential)]
        private class MinitorInfo
        {
            public int cbSize = Marshal.SizeOf(typeof(MinitorInfo));
            public Rect rcMonitor;
            public Rect rcWork;
            public int dwFlags;
        }

        private static MinitorInfo GetMonitorInfo(IntPtr hMonitor)
        {
            var mi = new MinitorInfo();
            _GetMonitorInfo(hMonitor, mi);
            return mi;
        }

        #endregion // Types
    }
}
